"""MMCretry.py

An M/M/c queue with retrials

Jobs arrive at random into a c-server queue with
exponential service-time distribution. Simulate to
determine the average  number and the average time
in the system. 

- c = Number of servers = 3
- arate = Arrival rate = 2.0
- stime = mean service time = 1.0

"""
from SimPy.Simulation import *
from random import expovariate,seed,weibullvariate
import time

## Model components ------------------------

def main():
    
    #LOOPS
    numberOfTrials = 500
    i = 0
    print "Simulating "+str(numberOfTrials)+" runs...\n"
    while i < numberOfTrials:
        i = i + 1;
        print "Running trial "+str(i)+"..."

        class Generator(Process):
            """ generates Jobs at random """

            def execute(self,maxNumber,arate,srate,rtime):
                ''' generate Jobs at exponential intervals '''
                for i in range(maxNumber):
                    L = Job("Job "+`i`)
                    activate(L,L.execute(srate),delay=0)
                    yield hold,self,expovariate(arate)
         

        class Job(Process):
            ''' Jobs request a gatekeeper and hold
                it for an exponential time '''
                
            NoInSystem=0
            NoInSvc=0
            def execute(self,srate):       
                arrTime=now()
                ntries = 1 ## everybody tries at least once
        #        self.trace("Hello World")
            #Loa shouldn't include yourself, hence doing this before incrementing.
                Loa.observe(Job.NoInSystem-Job.NoInSvc)
                mL.observe(Job.NoInSystem)
                mLs.observe(Job.NoInSvc)
                if (Job.NoInSvc >= c) :
                    pd.observe(1)
                else:
                    pd.observe(0)
        #   pd.observe(Job.NoInSystem >= c)
                Job.NoInSystem +=1
        #   ''' here is the retrial bit '''
                while (Job.NoInSvc >= c) :
                    ntries += 1
        # switch these next two lines (t=rtime and t=expovariate) to have
        # deterministic or exponential retrials, whichever line comes last.
        #            t = expovariate(rrate)
        #            t = weibullvariate(ascalepremult,bshape)
                    t = rtime
                    mWoeach.observe(t)
        #            self.trace("retrying...    ")
                    yield hold,self,t
        #   ''' done with the while loop--must be a free server'''
                mnt.observe(ntries)
                mWo.observe(now()-arrTime)
                yield request,self,server
        #        self.trace("At last    ")
        #        t = stime
        #        t =  1.0 / srate 
                t = expovariate(srate)
                mWs.observe(t)
                Job.NoInSvc +=1
                yield hold,self,t
                yield release,self,server
                Job.NoInSystem -=1
                Job.NoInSvc -=1
        # original MMC.py code also observes L, Lsvc on the way out
        # but that's not a nice clean PASTA way to do it,
        # unless you take the time-average using mL.timeAverage()
        #        mL.observe(Job.NoInSystem)
        #        mLs.observe(Job.NoInSvc)
                mW.observe(now()-arrTime)
        #        self.trace("Geronimo   ")
               
            def trace(self,message):
                FMT="%7.4f %6s %10s (%2d)"
                if TRACING:
                    print FMT%(now(),self.name,message,Job.NoInSystem)

        ## Experiment data -------------------------



        TRACING = True
        TRACING = False
        stime = 1.0     ## mean service time
        srate = 1.0/stime     ## mean service rate
        arate  = 9.0     ## mean arrival rate 
        c = 3           ## number of servers in M/M/c system
        c = 12           ## number of servers in M/M/c system

        ##for loop to try different rates should be done!!!
        #rrate = 0.1 ## the smallest we go in Matlab simulations
        #rrate = 0.01
        #rrate = 0.001
        #rrate = 2.0

        #my code (seeding a random number generator from 0.0 to 2.0

        myseed=random.uniform(24601,100000) ## an arbitrary number (a prisoner's number in Les Miserable)
        rrate = random.uniform(0.0, 2.0)

        rtime = 1.0/rrate     ## mean retrial time
        maxNumber= 400000 ## at rrate=0.01
        maxNumber= 4000000 ## at rrate = 0.001
        maxNumber= 40000 ## at rrate=0.1 or higher
        mL   =Monitor()  ## monitor for the number of jobs
        mW   =Monitor()  ## monitor for the time in system
        mWo  =Monitor()  ## monitor for the time in Orbit
        mnt  =Monitor()  ## monitor for # of tries
        mWs  =Monitor()  ## monitor for the generated service times
        pd   =Monitor() ## monitor for Pr(busy) upon outside arrival
        Loa  =Monitor() ## monitor for Lo upon outside arrival
        mLs  =Monitor() ## monitor for # in service
        mWoeach = Monitor() ## monitor for each orbit duration

       
        #seed(time.time()) ## seed for random numbers
        #seed(333555777) ## seed for random numbersr


        targetscv = 1.0

        #SCV    a=scale b=shape
                                 #1                         1                         1
        #targetscv=1.0
        if (targetscv == 1.0) :
            ascale=1.0
            bshape=1.0
                               #0.5          1.10132063443578          1.43552247479565
        #targetscv=0.5
        if (targetscv == 0.5) :
            ascale=1.10132063443578
            bshape=1.43552247479565
                 #0.333333333333333          1.12423402024472          1.79147384946114
        #targetscv=0.333333333333333
        # This is the only one where we have to worry about precision
        # (how many 3's the user specified)
        if (abs(targetscv - 1.0/3.0) <0.01) :
            ascale=1.12423402024472
            bshape=1.79147384946114
                              #0.25          1.12906339181181          2.10134905271944
        #targetscv=0.25
        if (targetscv == 0.25) :
            ascale=1.12906339181181
            bshape=2.10134905271944
                               #0.2           1.1282247738558          2.37977102271811
        #targetscv=0.2
        if (targetscv == 0.2) :
            ascale=1.1282247738558
            bshape=2.37977102271811
                               #0.1          1.11137554126024          3.50283931986704
        #targetscv=0.1
        if (targetscv == 0.1) :
            ascale=1.11137554126024
            bshape=3.50283931986704
                              #0.01          1.04303769567683          12.1533744625368
        #targetscv=0.01
        if (targetscv == 0.01) :
            ascale=1.04303769567683
            bshape=12.1533744625368
                            #0.0001          1.00448576470104          127.530115426386
        #targetscv = 0.0001
        if (targetscv == 0.0001) :
            ascale=1.00448576470104   
            bshape=127.530115426386
        #all with a mean of 1.  For other means, multiply "a" by desired mean.
        #python has alpha=scale, beta=shape: weibullvariate(self,alpha,beta)
        if (targetscv == 0.0000) :
            # technically, needs bshape=infinity to get Weibull with scv=0
            # We'll only runn this with tscv=0 if we've changed away
            # from using Weibull, so we'll just assign junk here:
            ascale=-999.0
            bshape=-999.0

        ascalepremult = ascale * 1.0 / rrate

        server=Resource(capacity=c,name='Gatekeeper')


        ## Model/Experiment ------------------------------
            

        seed(myseed)
        initialize()
        g = Generator('gen')
        activate(g,g.execute(maxNumber=maxNumber,
                     arate=arate, srate=srate,rtime=rtime))

        simulate(until=0.90*maxNumber/arate)


    ## Analysis/output -------------------------

        traffic = 1.0*arate*stime
        oscv = mWoeach.var() / mWoeach.mean()**2
        
        print "rho  c   trr orr tscv    oscv    seed    ncalls  Lo  Wo  ntries  pd"
        print "%02d %02d    %7.3f   %7.4f   %6.5f   %6.5f   %d  %d  %6.4f   %6.4f   %6.4f   %6.4f"%(traffic,c,rrate,1.0/mWoeach.mean(),targetscv,oscv,myseed, maxNumber, Loa.mean(), mWo.mean(), mnt.mean(), pd.mean())+"\n"

    print "Build Successful"
    
#initial main call 
if __name__ == "__main__":
        main()

        # would also be good to compute Pr(retry fails)
        # Let's see.  If everyone tried 1 time exactly, then everybody
        # got straight into service=no retrials happened at all
        # If everybody tried 2 times exactly, then everybody got bounced upon
        # initial arrival, then retried once, and got in, so Pr(retry fails)=0
        # If everybody tried 3 times exactly, then Pr(retry fails)=50%
        # So, Pr(retry fails) = (avg#tries-2)/(avg#tries-1) ???

        #print "%2d servers, %6.4f arrival rate,%6.4f mean service time,%6.4f rtime"%(c,arate,stime,rtime)
        #reltraffic = traffic/(1.0*c);
        #print "%6.4f traffic,%6.4f reltraffic"%(traffic,reltraffic)
        #print "If M/M/1, L=%g"%(reltraffic/(1-reltraffic))
        #print "Average number in the system is %6.4f"%(m.timeAverage(),)
        #print "Average time in the system is   %6.4f"%(mW.mean(),)
        #print "Actual average service-time is  %6.4f"%(mWs.mean(),)
        #print "Pr(delay upon arrival) is %6.4f"%(pd.mean(),)
        #print "Average number in service is %6.4f"%(mLs.mean(),)

